-- create parser with "happy -g -a -c HaskiParser.y"
{{-
Part of "Haski the Robot"
examination project for adv. funct. progr. 2004

    Copyright (C) 2004  Lars Otten

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-}

{-

  I M P O R T A N T  ! ! !
----------------------------
after creating the parser with happy make the following change
at the end of the definition of
"happyNewToken action sts stk (tk:tks)":
Change the line
	_ -> happyError tks
to
	_ -> happyError (tk:tks)

-}

module HaskiParser(
		haskiParser,	-- the actual parsing function
		Program(..),	-- the datatype for the commands
		Function(..),	-- represents a function definition
		Sensor(..),	-- contains the sensors of the robot
		LexE(..)	-- Monad for handling errors in parsing
		) where

import HaskiLexer
import ErrorMonad

}

%name parse
%tokentype { Token }
%monad { LexE }

%token
	'='		{ T _ _ TDefine}
	'>'		{ T _ _ TSeq }
	'('		{ T _ _ TBO }
	')'		{ T _ _ TBC }
	go		{ T _ _ TGo_forward }
	turn		{ T _ _ TTurn_right }
	take		{ T _ _ TTake_item }
	drop		{ T _ _ TDrop_item }
	idle		{ T _ _ TDo_nothing }
	while		{ T _ _ TWhile }
	it		{ T _ _ TIfThen }
	ite		{ T _ _ TIfThenElse }
	dowhile		{ T _ _ TDoWhile }
	'&'		{ T _ _ TAND }
	'|'		{ T _ _ TOR }
	'!'		{ T _ _ TNOT }
	front_free	{ T _ _ TFront_free }
	right_free	{ T _ _ TRight_free }
	left_free	{ T _ _ TLeft_free }
	is_carrying	{ T _ _ TIs_carrying }
	field_has_item	{ T _ _ TField_has_item }
	facing_up	{ T _ _ TFacing_up }
	facing_right	{ T _ _ TFacing_right }
	facing_down	{ T _ _ TFacing_down }
	facing_left	{ T _ _ TFacing_left }
	name		{ T _ _ (TName $$) }


%left '>'
%left '|'
%left '&'
%left '!'



%%

Main	:: { [Function] }
	: name '=' Func				{ [Def $1 $3] }
	| Main name '=' Func			{ (Def $2 $4) : $1 }

Func	:: { Program } 
	: Command				{ $1 }
	| '(' Func ')'				{ $2 }
	| Func '>' Func				{ Seq $1 $3 }

Command :: { Program }
	: go					{ Forward }
	| turn					{ Turn }
	| take					{ Take }
	| drop					{ Drop }
	| idle					{ Idle }
	| while Cond Command			{ While $2 $3 }
	| while Cond '('Func')'			{ While $2 $4 }
	| dowhile Command Cond			{ Seq $2 (While $3 $2) }
	| dowhile '('Func')' Cond		{ Seq $3 (While $5 $3) }
	| it Cond Command			{ IfThen $2 $3 }
	| it Cond '('Func')'			{ IfThen $2 $4 }
	| ite Cond Command Command		{ IfThenElse $2 $3 $4 }
	| ite Cond Command '('Func')'		{ IfThenElse $2 $3 $5 }
	| ite Cond '('Func')' Command		{ IfThenElse $2 $4 $6 }
	| ite Cond '('Func')' '('Func')'	{ IfThenElse $2 $4 $7 }
	| name					{ Call $1 }

Cond	:: { Sensor }
	: Bool					{ $1 }
	| '!' Bool				{ Not $2 }
	| '!' '(' Cond2 ')'			{ Not $3 }
	| '(' Cond2 ')'				{ $2 }


Cond2	:: { Sensor }
	: Cond2 '&' Cond2			{ $1 :& $3 }
	| Cond2 '|' Cond2			{ $1 :| $3 }
	| '!' Cond				{ Not $2 }
	| Bool					{ $1 }

Bool 	:: { Sensor }
	: front_free				{ Front_free }
	| right_free				{ Right_free }
	| left_free				{ Left_free }
	| is_carrying				{ Is_carrying }
	| field_has_item			{ Field_has_item }
	| facing_up				{ Facing_up }
	| facing_right				{ Facing_right }
	| facing_down				{ Facing_down }
	| facing_left				{ Facing_left }


{

happyError :: [Token] -> LexE a
happyError ((T p s _):ts) = Failed ("Line " ++ getLineNo p
               ++ ": Parse error on string \"" ++ s ++ "\"")
  where getLineNo (AlexPn _ l _) = show l
happyError [] = Failed "Parse error on end of file"

-------------------------------------------
-- output data type

data Sensor
 = Front_free
 | Right_free
 | Left_free
 | Field_has_item
 | Is_carrying
 | Facing_up
 | Facing_right
 | Facing_down
 | Facing_left
 | Sensor :& Sensor -- AND
 | Sensor :| Sensor -- OR
 | Not Sensor       -- Negation

instance Show Sensor where
 show Front_free = "front_free"
 show Right_free = "Right_free"
 show Left_free = "Left_free"
 show Field_has_item = "field_has_item"
 show Is_carrying = "is_carrying"
 show Facing_up = "facing_up"
 show Facing_right = "facing_right"
 show Facing_down = "facing_down"
 show Facing_left = "facing_left"
 show (Not p) = "Not " ++ show p
 show (p :& q) = "(" ++ show p ++ " :& " ++ show q ++ ")"
 show (p :| q) = "(" ++ show p ++ " :| " ++ show q ++ ")"

type Main
 = [Function]

data Function
 = Def String Program
    deriving Show

data Program
 = IfThen Sensor Program
 | IfThenElse Sensor Program Program
 | While Sensor Program
 | Forward
 | Turn
 | Take
 | Drop
 | Idle
 | Seq Program Program
 | End
 | Call String
    deriving Show

-- main parsing function
haskiParser input = lexer input >>= parse

}
